/**
 * Copyright Copyright 2007-8 Simon Andrews
 *
 *    This file is part of StackMeasure.
 *
 *    StackMeasure is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    StackMeasure is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with StackMeasure; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
import java.awt.Color;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;


public class GroupCollection {

	//TODO:  There is a problem with this model whereby it assumes that all features are
	// somewhat spherical such that they can be built up from a top down view.  However
	// there are irregularly shaped features which don't fit into this pattern.  We need
	// to find some what to post merge spatial groups together, but this would mean
	// redefining how the outer limits of a group are modelled otherwise you just get
	// the outside of the combined set (which would be too big).
	
	private Vector groups;
	private StackMeasurePreferences prefs;
	private FileList files;
	
	public GroupCollection (StackMeasurePreferences prefs,FileList files) {
		this.prefs = prefs;
		groups = new Vector();
		this.files = files;
		
		// For consistency when processing multiple sets of files we'll
		// reset the colour index of StackMeasure file.  This means a set
		// of files will always get the same colours for the same channels.
		StackMeasureFile.resetColourIndexes();
	}
	
	public FileList getFiles () {
		return files;
	}
		
	public void addPixel (Pixel p) {
		// Try to put this pixel into one of the existing groups
		Enumeration e = groups.elements();
		while (e.hasMoreElements()) {
			if (((SpatialGroup)e.nextElement()).addPixel(p)) {
				return;
			}
		}
		// If we get here then none of the existing groups wanted it
		// so let's start a new one
		SpatialGroup g = new SpatialGroup(p,this);
		groups.add(g);
	}
	
	public void mergeSpatialGroups (StackMeasureProcessorListener [] listeners) {
		boolean mergedSomething = true;
		
		int passNumber = 0;
		
		while (mergedSomething) {
			passNumber++;
			for (int l=0;l<listeners.length;l++) {
				listeners[l].progressUpdate(1, 1, "Merging Groups (pass "+passNumber+")");
			}
			
			// To save doing the same comparison lots of times we're going
			// to keep a record of deleted (merged) groups and just skip
			// these when we got through the list.
			Hashtable deletedGroups = new Hashtable();
						
			mergedSomething = false;
			SpatialGroup [] sg = getSpatialGroups();
			for (int s1=0;s1<sg.length;s1++) {
				if (deletedGroups.containsKey(sg[s1])) continue;
				for (int s2=s1+1;s2<sg.length;s2++) {
					if (deletedGroups.containsKey(sg[s2])) continue;
					if (sg[s1].mergeWith(sg[s2])) {
						deletedGroups.put(sg[s2],"");
						groups.remove(sg[s2]);
						mergedSomething = true;
					}					
				}
			}
			// We can now clear the deleted groups list
			deletedGroups.clear();
		}
		
	}
		
	public int getGroupCount () {
		return groups.size();
	}
	
	public StackMeasurePreferences getPrefs () {
		return prefs;
	}
		
	public SpatialGroup [] getSpatialGroups() {
		
		final Vector v = new Vector();
		
		Enumeration e = groups.elements();
		while (e.hasMoreElements()) {
			SpatialGroup s = (SpatialGroup)e.nextElement();
			if (s.getPixelCount()>=prefs.minPixels()) {
				v.add(s);
			}
		}
		
		SpatialGroup [] cg = (SpatialGroup [])v.toArray(new SpatialGroup[0]);
		
		Arrays.sort(cg);
		
 		return cg; 
	}

	public ColourGroup [] getColourGroups() {
		
		SpatialGroup [] sg = getSpatialGroups();

		final Vector v = new Vector();
		
		SPATIALS: for (int i=0;i<sg.length;i++) {
			Enumeration e2 = v.elements();
			while (e2.hasMoreElements()) {
				ColourGroup cg = (ColourGroup)e2.nextElement();
				if (cg.addSpatialGroup(sg[i])) {
					continue SPATIALS;
				}
			}
			
			// If we get here we need to start a new group
			ColourGroup cg = new ColourGroup(sg[i],this);
		
			v.add(cg);
		}
		
 		return (ColourGroup [])v.toArray(new ColourGroup[0]);
	}

	public void saveSpatialHTMLTable(File file) throws IOException {

		PrintWriter p = new PrintWriter(new FileWriter(file));

		//Start off the html file
		p.println("<html>");
		p.println("<head>");
		p.println("<title>Spatial Groups</title>");
		p.println("<style type=\"text/css\">");
		p.println("td, th {");
		p.println("border-style:solid;");
		p.println("border-width: 1px;");
		p.println("}");
		p.println("</style>");
		p.println("</head>");
		p.println("<body>");
		p.println("<h1>Spatial Groups</h1>");
		p.println("<table>");

		// Start with the headers
		p.println("<tr>");
		String [] headers = new String [] {"Colour","Pixel Count","Volume","Start Slice","End Slice","Hue"};
		for (int i=0;i<headers.length;i++) {
			p.println("<th>"+headers[i]+"</th>");
		}
		p.println("</tr>");

		//Now work our way through the table
		Enumeration e = groups.elements();
		while (e.hasMoreElements()) {
			SpatialGroup g = (SpatialGroup)e.nextElement();
			
			p.println("<tr>");

			//Colour
			p.println("<td style=\"background:"+getHexString(g.getBrightestColor())+"\"><pre>"+g.getFile().name()+"</pre></td>");
						
			// Pixel Count
			p.println("<td>"+g.getPixelCount()+"</td>");

			// Volume
			p.println("<td>"+g.getVolume()+"</td>");
			
			// Start Slice
			p.println("<td>"+g.firstSlice()+"</td>");
			
			// End Slice
			p.println("<td>"+g.lastSlice()+"</td>");
			
			// Hue
			p.println("<td>"+g.getAverageHue()+"</td>");

			
			p.println("</tr>");
		}

		p.println("</table></body></html>");

		p.close();
	}

	public void saveSpatialTextTable(File file) throws IOException {

		PrintWriter p = new PrintWriter(new FileWriter(file));


		// Start with the headers
		String [] headers = new String [] {"Colour","Pixel Count","Volume","Start Slice","End Slice","Hue"};
		for (int i=0;i<headers.length;i++) {
			p.print(headers[i]+"\t");
		}
		p.print("\n");

		//Now work our way through the table
		Enumeration e = groups.elements();
		while (e.hasMoreElements()) {
			SpatialGroup g = (SpatialGroup)e.nextElement();
			
			//Colour
			p.print(g.getFile().name()+"\t");
						
			// Pixel Count
			p.print(g.getPixelCount()+"\t");

			// Volume
			p.print(g.getVolume()+"\t");
			
			// Start Slice
			p.print(g.firstSlice()+"\t");
			
			// End Slice
			p.print(g.lastSlice()+"\t");
			
			// Hue
			p.print(g.getAverageHue()+"\t");

			
			p.print("\n");
		}

		p.close();
	}

	
	private String getHexString (Color c) {
		StringBuffer b = new StringBuffer("#");

		int i = c.getRed();
		String s = Integer.toHexString(i);
		if (s.length() == 1) b.append("0");
		b.append(s);

		i = c.getGreen();
		s = Integer.toHexString(i);
		if (s.length() == 1) b.append("0");
		b.append(s);

		i = c.getBlue();
		s = Integer.toHexString(i);
		if (s.length() == 1) b.append("0");
		b.append(s);

		return b.toString();
	}
}
