/**
 * Copyright Copyright 2007-8 Simon Andrews
 *
 *    This file is part of StackMeasure.
 *
 *    StackMeasure is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    StackMeasure is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with StackMeasure; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
import ij.ImageListener;
import ij.ImagePlus;
import ij.gui.StackWindow;

import java.awt.BorderLayout;
import java.awt.Color;
import java.util.Enumeration;
import java.util.Vector;

import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.table.AbstractTableModel;


public class SpatialResultsViewer extends JPanel implements ImageListener, ListSelectionListener {
	
	private GroupCollection collection;
	private SpatialGroup [] groups;
	private SpatialGroup selectedGroup = null;
	private JTable table;
	private Vector listeners = new Vector();
	private ResultsModel model;
	
	public SpatialResultsViewer (GroupCollection collection) {
		this.collection = collection;
		ImagePlus.addImageListener(this);
		groups = collection.getSpatialGroups();
		setLayout(new BorderLayout());
		model = new ResultsModel();
		table = new JTable(model);
		table.getColumnModel().getColumn(0).setCellRenderer(new ColorCellRenderer());
		table.setDefaultRenderer(Color.class, new ColorCellRenderer());
		table.getSelectionModel().addListSelectionListener(this);
		add(new JScrollPane(table),BorderLayout.CENTER);
	}
	
	public void addListener (SpatialResultsListener l) {
		listeners.add(l);
	}
	
	public void removeListener (SpatialResultsListener l) {
		if (listeners.contains(l)) {
			listeners.remove(l);
		}
	}
	
	private void selectRegion () {

		if (selectedGroup != null) {
			int [] i = selectedGroup.getExtentForSlice(collection.getFiles().getMergedImage().getCurrentSlice());
			
			if (i == null) {
//				System.out.println("No ROI in this slice ("+collection.getImagePlus().getCurrentSlice()+")");
				collection.getFiles().getMergedImage().setRoi(0,0,0,0);
			}
			else {
//				System.out.println("Setting ROI to "+i[0]+","+i[1]+","+i[2]+","+i[3]);
				collection.getFiles().getMergedImage().setRoi(i[0]-1,i[1]-1,(i[2]-i[0])+2,(i[3]-i[1])+2);
			}
		}
	}
	
	private class ResultsModel extends AbstractTableModel {
		
		private String [] headers = new String [] {"Colour","Pixel Count","Volume","Start Slice","End Slice","Hue"};
		private Class [] classes = new Class [] {String.class,Integer.class,Double.class,Integer.class,Integer.class,Integer.class};
		
		public int getColumnCount() {
			return 6;
		}

		public int getRowCount() {
			return groups.length;
		}

		public String getColumnName (int col) {
			return headers[col];
		}
		
		public Class getColumnClass (int col) {
			return classes[col];
		}
		
		public Object getValueAt(int row, int col) {
			switch (col) {
				case 0:
					return groups[row].getBrightestColor();
				case 1:
					return new Integer(groups[row].getPixelCount());
				case 2:
					return new Double (groups[row].getVolume());
				case 3:
					return new Integer(groups[row].firstSlice());
				case 4:
					return new Integer(groups[row].lastSlice());
				case 5:
					return new Integer(groups[row].getAverageHue());
			}
			return null;
		}
		
	}

	public void imageClosed(ImagePlus ip) {
	}
	public void imageOpened(ImagePlus arg0) {}

	public void imageUpdated(ImagePlus ip) {
		if (ip.equals(collection.getFiles().getMergedImage())) {
			selectRegion();
		}
	}

	public void valueChanged(ListSelectionEvent lse) {
		// Check how many selections have been made
		if (table.getSelectedRows().length == 1) {
			selectedGroup = groups[table.getSelectedRow()];
			//TODO:  This can fail if the composite image isn't visible
			((StackWindow)collection.getFiles().getMergedImage().getWindow()).showSlice(selectedGroup.middleSlice());

			// Tell any listeners what we've done
			Enumeration e = listeners.elements();
			while (e.hasMoreElements()) {
				((SpatialResultsListener)e.nextElement()).spatialGroupSelected(selectedGroup);
			}
		}
		else {
			// Tell any listeners what we've done
			Enumeration e = listeners.elements();
			while (e.hasMoreElements()) {
				((SpatialResultsListener)e.nextElement()).clearSelections();
			}
			selectedGroup = null;
		}
		selectRegion();
	}

}
