/**
 * Copyright Copyright 2007-8 Simon Andrews
 *
 *    This file is part of StackMeasure.
 *
 *    StackMeasure is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    StackMeasure is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with StackMeasure; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
import java.util.Enumeration;
import java.util.Vector;
import ij.ImageStack;


public class StackMeasureProcessor implements Runnable {

	private FileList fileList;
	private Vector listeners;
	private GroupCollection collection;
	
	public StackMeasureProcessor (GroupCollection collection) {
		if (collection == null) {
			throw new NullPointerException("Collection can't be null");
		}
		this.collection = collection;
		listeners = new Vector();
	}
	
	public void addSparkMeasureProcessorListener (StackMeasureProcessorListener l) {
		if (! listeners.contains(l)) {
			listeners.add(l);
		}
	}

	public void removeSparkMeasureProcessorListener (StackMeasureProcessorListener l) {
		if (listeners.contains(l)) {
			listeners.remove(l);
		}
	}

	
	public void startProcessing (FileList fileList) {
		this.fileList = fileList;
		Thread t = new Thread(this);
		t.start();
	}
	
	public void run() {
		try {
			System.err.println("Starting processing");


			StackMeasureFile [] files = fileList.getFiles();
			
			for (int f=0;f<files.length;f++) {

				if (files[f].isRGB()) {
					processRGBFile(files[f]);
				}
				else {
					processBWFile (files[f]);
				}
			}

			Enumeration e = listeners.elements();
			while (e.hasMoreElements()) {
				((StackMeasureProcessorListener)e.nextElement()).progressUpdate(10, 10, "Merging groups");
			}
			try {
				Thread.sleep(50);
			} catch (InterruptedException e1) {}

			collection.mergeSpatialGroups((StackMeasureProcessorListener [])listeners.toArray(new StackMeasureProcessorListener[0]));

			e = listeners.elements();
			while (e.hasMoreElements()) {
				((StackMeasureProcessorListener)e.nextElement()).progressUpdate(10, 10, "Creating Composite Image");
			}
			
			
			// If we get here then we've finished
			e = listeners.elements();
			while (e.hasMoreElements()) {
				((StackMeasureProcessorListener)e.nextElement()).processingComplete();
			}
			listeners.clear();

		}
		catch (Exception ex) {
			Enumeration e = listeners.elements();
			while (e.hasMoreElements()) {
				((StackMeasureProcessorListener)e.nextElement()).exceptionReceived(ex);
			}
			listeners.clear();
		}
	}
	
	private void processRGBFile (StackMeasureFile file) {
		
		ImageStack stack = file.getImagePlus().getImageStack();
		
		int width = stack.getWidth();
		int height = stack.getHeight();

		for (int s=1;s<=stack.getSize();s++) {
//			System.err.println("Doing slice "+s);
			Enumeration e = listeners.elements();
			while (e.hasMoreElements()) {
				((StackMeasureProcessorListener)e.nextElement()).progressUpdate(s, stack.getSize(), "Processing slice "+s+" of "+stack.getSize()+" in "+file.name());
			}
			try {
				Thread.sleep(50);
			} catch (InterruptedException e1) {}


			// Now we need to get the data for this slice
			int [] pixelData = (int [])stack.getPixels(s);

			for (int w=0;w<width;w++) {
				for (int h=0;h<height;h++) {
					// We need to split the integer value into its RGB components
					byte [] rgb = Pixel.intToByteArray(pixelData[w+(h*width)]);
					Pixel p = new Pixel(file,rgb[0],rgb[1],rgb[2],s,w,h);

					// Check if this pixel is bright enough to use
					if (p.l() < collection.getPrefs().minIntensity()) {
						continue;
					}

//					System.out.println("RGB was "+rgb[0]+":"+rgb[1]+":"+rgb[2]+" at slice "+s+" pixel "+w+h);


					// OK so this is a usable pixel.  Lets put it into the
					// collection
					collection.addPixel(p);
				}
			}

		}
	}
	
	private void processBWFile (StackMeasureFile file) {
		ImageStack stack = file.getImagePlus().getImageStack();
		
		int width = stack.getWidth();
		int height = stack.getHeight();

		for (int s=1;s<=stack.getSize();s++) {
//			System.err.println("Doing slice "+s);
			Enumeration e = listeners.elements();
			while (e.hasMoreElements()) {
				((StackMeasureProcessorListener)e.nextElement()).progressUpdate(s, stack.getSize(), "Processing slice "+s+" of "+stack.getSize()+" in "+file.name());
			}
			try {
				Thread.sleep(50);
			} catch (InterruptedException e1) {}


			// Now we need to get the data for this slice
			Object o = stack.getPixels(s);
			
			int [] pixelData;
			int intensityCutoff;
			
			if (o instanceof byte []) {
				byte [] a = (byte [])o;
				// It's an 8-bit greyscale image
				intensityCutoff = (int)(collection.getPrefs().minIntensity() * 2.56);
				pixelData = new int [a.length];
				for (int i=0;i<a.length;i++) {
					pixelData[i] = a[i] & 0xff;
				}
				
				
			}
			else if (o instanceof short []) {
				short [] a = (short [])o;
				// It's a 16 bit greyscale image
				intensityCutoff = (int)(collection.getPrefs().minIntensity() * 655.36);
				pixelData = new int [a.length];
				for (int i=0;i<a.length;i++) {
					pixelData[i] = (int)a[i];
				}
			}
			
			else {
				System.err.println("Unknown greyscale file type which produces "+o.getClass());
				return;
			}
			

			for (int w=0;w<width;w++) {
				for (int h=0;h<height;h++) {
					// We need to split the integer value into its RGB components
					int intensity = pixelData[w+(h*width)];

					// Check if this pixel is bright enough to use
					if (intensity < intensityCutoff) {
						continue;
					}
					Pixel p = new Pixel(file,s,w,h);

//					System.out.println("RGB was "+rgb[0]+":"+rgb[1]+":"+rgb[2]+" at slice "+s+" pixel "+w+h);


					// OK so this is a usable pixel.  Lets put it into the
					// collection
					collection.addPixel(p);
				}
			}

		}
	}
		
}
