/**
 * Copyright Copyright 2007-8 Simon Andrews
 *
 *    This file is part of StackMeasure.
 *
 *    StackMeasure is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    StackMeasure is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with StackMeasure; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.GraphicsConfiguration;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.util.Hashtable;

import ij.ImageListener;
import ij.ImagePlus;

import javax.media.j3d.Appearance;
import javax.media.j3d.BoundingSphere;
import javax.media.j3d.BranchGroup;
import javax.media.j3d.Canvas3D;
import javax.media.j3d.ColoringAttributes;
import javax.media.j3d.LineArray;
import javax.media.j3d.Node;
import javax.media.j3d.Shape3D;
import javax.media.j3d.Transform3D;
import javax.media.j3d.TransformGroup;
import javax.media.j3d.TransparencyAttributes;
import javax.swing.JPanel;
import javax.vecmath.Color3f;
import javax.vecmath.Point3f;
import javax.vecmath.Vector3f;

import com.sun.j3d.utils.behaviors.mouse.MouseRotate;
import com.sun.j3d.utils.geometry.Box;
import com.sun.j3d.utils.universe.SimpleUniverse;


public class VolumeDisplay extends JPanel implements MouseWheelListener, ImageListener, SpatialResultsListener, DistanceSelectionListener {

	private final GroupCollection collection;
	
	private float graphScale;
	private float xOffset;
	private float yOffset;
	private float zOffset;
	private float pixelGraphSize;
	private float stackGraphSize;
	private boolean scalesCalculated = false;
	private int currentSlice;
	private TransformGroup sliceGroup;
	private TransformGroup wholeScene = null;
	private Hashtable spatialGroupAppearences = new Hashtable();
	
	
	// Temp variables used for highlighting
	private SpatialGroup highlightedGroup = null;
	private Node distanceMeasure = null;
	private final Color3f highlightColour = new Color3f(Color.YELLOW);
	
	public VolumeDisplay (GroupCollection collection) {
		addMouseWheelListener(this);
		this.collection = collection;
		setMinimumSize(new Dimension(0,0));
		currentSlice = 0;
		
		ImagePlus.addImageListener(this);
		
		GraphicsConfiguration config = SimpleUniverse.getPreferredConfiguration();

		Canvas3D canvas = new Canvas3D(config);
		setLayout(new BorderLayout());
		add(canvas,BorderLayout.CENTER);

		BranchGroup scene = createSceneGraph();

		SimpleUniverse simpleU = new SimpleUniverse(canvas);

		simpleU.getViewingPlatform().setNominalViewingTransform();

		simpleU.addBranchGraph(scene);
		
	}
		
	protected void highlightGroup (SpatialGroup g) {

		removeHighlight();
		removeDistanceMeasure();
		
		if (spatialGroupAppearences.containsKey(g)) {
			Appearance a = (Appearance)spatialGroupAppearences.get(g);
			a.getColoringAttributes().setColor(highlightColour);
			highlightedGroup = g;
		}
	}
	
	protected void removeHighlight () {
		if (highlightedGroup != null) {
			Appearance a = (Appearance)spatialGroupAppearences.get(highlightedGroup);
			a.getColoringAttributes().setColor(new Color3f(highlightedGroup.getBrightestColor()));			
			highlightedGroup = null;
		}
	}
	
	protected void removeDistanceMeasure () {
		if (distanceMeasure != null) {
			wholeScene.removeChild(distanceMeasure);
			distanceMeasure = null;
		}
	}
	
	protected void showDistanceMeasure (DistanceMeasure d) {
		// We need to make up a line showing the distance measure
	
		removeDistanceMeasure();
		removeHighlight();
		
		// Check we're not being called too early.
		if (wholeScene == null) return;
		
		BranchGroup b = new BranchGroup();
		b.setCapability(BranchGroup.ALLOW_DETACH);
		
		Point3f [] p = new Point3f [2];
		p[0] = new Point3f(getBoxCoordinates(d.firstPixel()));
		p[1] = new Point3f(getBoxCoordinates(d.secondPixel()));
		LineArray la = new LineArray(2,LineArray.COORDINATES);
		la.setCoordinates(0, p);
		Shape3D line = new Shape3D(la,new Appearance());
		ColoringAttributes ca = new ColoringAttributes();
		ca.setColor(highlightColour);
		line.getAppearance().setColoringAttributes(ca);
		b.addChild(line);
		wholeScene.addChild(b);
		
		distanceMeasure =  b;
		
	}
	
	protected void setHighlightedSlice () {
		
		// If people are a bit quick off the mark this could
		// get called before the constructor has finished.
		if (sliceGroup == null) return;
		
//		currentSlice = collection.getImagePlus().getCurrentSlice();
//		System.out.println("Moving plane to "+currentSlice);
		
		Transform3D t = new Transform3D();
		float [] f = getBoxCoordinates(0,0,currentSlice);
		// We have to subtract half a stack size because of the
		// whole "measure out from the middle" thing of Java3D
		
		t.set(new Vector3f(0,0,f[2]-(stackGraphSize/2)));

		sliceGroup.setTransform(t);
		
	}
	
	protected BranchGroup createSceneGraph() {

		// Make a root for the whole scene
		BranchGroup root = new BranchGroup();

		// Have a master transform group which can be rotated with
		// the mouse.  All other features will fall below this
		wholeScene = new TransformGroup();
		wholeScene.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
		wholeScene.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
		wholeScene.setCapability(TransformGroup.ALLOW_CHILDREN_WRITE);
		wholeScene.setCapability(TransformGroup.ALLOW_CHILDREN_READ);
		wholeScene.setCapability(TransformGroup.ALLOW_CHILDREN_EXTEND);

		root.addChild(wholeScene);
		
		
		// Now iterate through the spatial groups drawing each of them
		// in an appropriate colour.
		SpatialGroup [] groups = collection.getSpatialGroups();
		
		for (int i=0;i<groups.length;i++) {
			createGroupGraph(groups[i],wholeScene);
		}
		
		MouseRotate myMouseRotate = new MouseRotate();
		myMouseRotate.setTransformGroup(wholeScene);
		myMouseRotate.setSchedulingBounds(new BoundingSphere());
		root.addChild(myMouseRotate);

//		currentSlice = collection.getImagePlus().getCurrentSlice();

		// We need to add a transparent plane to show which slice
		// we're currently looking at
		
		// The plane is just a box with a 1U thickness
		Appearance a = new Appearance();
		ColoringAttributes ca = new ColoringAttributes();
		ca.setColor(new Color3f(Color.WHITE));
		a.setColoringAttributes(ca);
		
		// Make the slice transparent
		TransparencyAttributes ta = new TransparencyAttributes();
		ta.setTransparencyMode(TransparencyAttributes.BLENDED);
		ta.setTransparency(0.8f);
		a.setTransparencyAttributes(ta);
		
		Transform3D t = new Transform3D();
		float [] f = getBoxCoordinates(0,0,currentSlice);
		t.set(new Vector3f(0,0,f[2]-(stackGraphSize/2)));
		sliceGroup = new TransformGroup(t);
		sliceGroup.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
		sliceGroup.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
		Box b = new Box(pixelGraphSize*collection.getFiles().getWidth(),pixelGraphSize*collection.getFiles().getHeight(),stackGraphSize,a);
//		Box b = new Box(0.1f,0.1f,0.1f,a);
		sliceGroup.addChild(b);
		wholeScene.addChild(sliceGroup);
		
//		root.compile();

		addBoundingBox(wholeScene);
		
		return root;
	}
	
	protected void addBoundingBox (TransformGroup parent) {
		// Add a very transparent box which delimits the whole scene
		// The plane is just a box with a 1U thickness
		Appearance a = new Appearance();

		// Make the slice transparent
		TransparencyAttributes ta = new TransparencyAttributes();
		ta.setTransparencyMode(TransparencyAttributes.BLENDED);
		ta.setTransparency(0.9f);
		a.setTransparencyAttributes(ta);

		ColoringAttributes ca = new ColoringAttributes();
		ca.setColor(new Color3f(Color.WHITE));
		a.setColoringAttributes(ca);

		Transform3D t = new Transform3D();
		t.set(new Vector3f(0.0f,0.0f,0.0f));
		TransformGroup boundingGroup = new TransformGroup(t);
		Box b = new Box(pixelGraphSize*collection.getFiles().getWidth(),pixelGraphSize*collection.getFiles().getHeight(),stackGraphSize*collection.getFiles().getSlices(),a);
		boundingGroup.addChild(b);
		parent.addChild(boundingGroup);

	}
	
	protected void createGroupGraph (SpatialGroup g, TransformGroup parent) {
		
//		System.out.println("Creating graph for spatial group");
		
		// Set up the appearance of this group
		Appearance a = new Appearance();
		a.setCapability(Appearance.ALLOW_COLORING_ATTRIBUTES_READ);
		a.setCapability(Appearance.ALLOW_COLORING_ATTRIBUTES_WRITE);
		
		ColoringAttributes ca = new ColoringAttributes();
		ca.setColor(new Color3f(g.getBrightestColor()));
		a.setColoringAttributes(ca);
		
		ca.setCapability(ColoringAttributes.ALLOW_COLOR_WRITE);
		
		// We're going to keep hold of the appearance so we can change it
		// later if we need to.
		spatialGroupAppearences.put(g, a);
		
		
		Pixel [] pixels = g.getPixels();
		
		// This can prove horrendously slow on big objects, so we're going
		// to limit each object to 200 pixels.  If it's more than that we'll
		// only take enough to get to that level.
		
		int skipValue = pixels.length/200;
		if (skipValue<1) skipValue = 1;
		
		for (int p=0;p<pixels.length;p++) {
			if (p%skipValue != 0) continue;
			Transform3D t = new Transform3D();
			t.set(new Vector3f(getBoxCoordinates(pixels[p])));
			TransformGroup tg = new TransformGroup(t);
			// We make the boxes bigger than the actual voxels to make a mode solid looking display
			Box b = new Box(pixelGraphSize*1.5f,pixelGraphSize*1.5f,stackGraphSize*1.5f,a);
			tg.addChild(b);
			parent.addChild(tg);
		}
	}
	
	protected float [] getBoxCoordinates (Pixel p) {
		return getBoxCoordinates(p.x(),p.y(),p.slice());
	}
	protected float [] getBoxCoordinates (int x, int y, int slice) {
		
		// Our standard view is to look down on the top of the stack.
		// We might change this to a side on view later, but for now
		// that's how it is.
		
		float [] xyz = new float[3];
		
		// The whole scene runs from -1 to +1.  We want the visible
		// volume to be -0.5 to +0.5.
		
		// Since the scene is scaled we need to convert pixel/slice
		// positions into real positions and then convert these into
		// scene graph positions.
		
		if (!scalesCalculated) {
//			System.out.println("Calculating scales");
			// We need to work out the longest axis (in um)
			float xLen = collection.getFiles().getWidth() * (float)collection.getPrefs().pixelDistance();
			float yLen = collection.getFiles().getHeight() * (float)collection.getPrefs().pixelDistance();
			float zLen = collection.getFiles().getSlices() * (float)collection.getPrefs().stackDistance();
			
			graphScale = xLen;
			if (yLen > graphScale) graphScale = yLen;
			if (zLen > graphScale) graphScale = zLen;
						
//			System.out.println("Graph scale is "+graphScale);
			
			// To convert to graph units we simply divide by the longest
			// axis length.
			
			// We also need to know the offset to use for each axis so that
			// the whole box is centered.  This will be -0.5 for the longest
			// axis, and somewhere between -0.5 and 0 for the others.
			xOffset = (xLen/graphScale) * -0.5f;
			yOffset = (yLen/graphScale) * -0.5f;
			zOffset = (zLen/graphScale) * -0.5f;
			
//			System.out.println ("Offsets are "+xOffset+":"+yOffset+":"+zOffset);
			
			// We will also need to know the dimensions of a pixel and a stack for
			// constructing the boxes
			pixelGraphSize = (float)collection.getPrefs().pixelDistance()/graphScale;
			stackGraphSize = (float)collection.getPrefs().stackDistance()/graphScale;
			
			// It appears that for boxes the measures radiate out from the central
			// point in both directions along the axis.  We therefore need measures
			// which are half what we might have expected.
			pixelGraphSize = pixelGraphSize/2;
			stackGraphSize = stackGraphSize/2;
		
//			System.out.println("Standard Box size is "+pixelGraphSize+","+pixelGraphSize+","+stackGraphSize);
			
			scalesCalculated = true;
		}
		
		// We can now use the existing scales to work out the extent of the Box
		xyz[0] = xOffset + ((x*(float)collection.getPrefs().pixelDistance())/graphScale);
		
		// I'd forgotten that the y-axis counts from the bottom not the top (why??!)
		// hence the 0- in this bit
		xyz[1] = 0 - (yOffset + ((y*(float)collection.getPrefs().pixelDistance())/graphScale));

		xyz[2] = zOffset + ((slice*(float)collection.getPrefs().stackDistance())/graphScale);
		
//		System.out.println("Position of "+p.x()+","+p.y()+","+p.slice()+" is "+xyz[0]+","+xyz[1]+","+xyz[2]);
		
		return xyz;
		
	}


	public void mouseWheelMoved(MouseWheelEvent e) {
		int w = e.getWheelRotation();
		int f = currentSlice+w;
		if (f<1) f=1;
		if (f>collection.getFiles().getSlices()) f=collection.getFiles().getSlices();

		currentSlice = f;
		collection.getFiles().getMergedImage().setSlice(currentSlice);
	}

	// Methods from SpatialResultsListener
	public void clearSelections() {
		removeHighlight();
		removeDistanceMeasure();
	}

	public void spatialGroupSelected(SpatialGroup g) {
		highlightGroup(g);
	}

	public void distanceSelected(DistanceMeasure d) {
		showDistanceMeasure(d);
	}

	public void imageClosed(ImagePlus arg0) {
	}

	public void imageOpened(ImagePlus arg0) {
	}

	public void imageUpdated(ImagePlus ip) {
		if (ip == collection.getFiles().getMergedImage()) {
			currentSlice = ip.getCurrentSlice();
			setHighlightedSlice();
		}
	}
	
}
